/*
 * MPC8260 FCC Fast Ethernet Driver
 *
 * Copyright (c) 2002 Italtel s.p.A   Enzo Figini (enzo.figini@italtel.it)
 *
 * Copyright (c) 2000 MontaVista Software, Inc.   Dan Malek (dmalek@jlc.net)
 *
 * (C) Copyright 2000 Sysgo Real-Time Solutions, GmbH <www.elinos.com>
 * Marius Groeger <mgroeger@sysgo.de>
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */
#include <common.h>
#include <asm/cpm_8260.h>
#include <mpc8260.h>
#include <net.h>
#include <command.h>
#include <config.h>
#include <malloc.h>
#include <miiphy.h>
#include <mii_phy.h>
#ifdef CONFIG_RINV
 #include <rinvLib.h> 
#endif

#if defined(FEC_ENET) && (CONFIG_COMMANDS & CFG_CMD_NET)
/*
#define DEBUG
*/
#define XMK_STR(x)	#x
#define MK_STR(x)	XMK_STR(x)

/* Maximum input DMA size.  Must be a should(?) be a multiple of 4. */
#define PKT_MAXDMA_SIZE         1520
/* The FCC stores dest/src/type, data, and checksum for receive packets. */
#define PKT_MAXBUF_SIZE         1518
#define PKT_MINBUF_SIZE         64
/* Maximum input buffer size.  Must be a multiple of 32. */
#define PKT_MAXBLR_SIZE         1536

#define TOUT_LOOP 1000000
#define	NUM_FCC		3
#define TX_BUF_CNT 	2

#define	DEV_FCC_NR	(int)(dev->priv)

#ifdef __GNUC__
static char txbuf[TX_BUF_CNT][PKT_MAXBLR_SIZE] __attribute__ ((aligned(8)));
#else
#error "txbuf must be 64-bit aligned"
#endif

/*
 * FCC Ethernet Tx and Rx buffer descriptors.
 * Provide for Double Buffering
 * Note: PKTBUFSRX is defined in net.h
 */
typedef volatile struct rtxbd {
    cbd_t rxbd[PKTBUFSRX];
    cbd_t txbd[TX_BUF_CNT];
} RTXBD;

typedef struct ethDevInfo
	{
	int	fccNr;
	} ETHDEVINFO;
	
/*  Good news: the FCC supports external BDs! */
#ifdef __GNUC__
static RTXBD rtx __attribute__ ((aligned(8)));
#else
#error "rtx must be 64-bit aligned"
#endif

static uint rxIdx[NUM_FCC];	 /* index of the current RX buffer */
static uint txIdx[NUM_FCC];	 /* index of the current TX buffer */
static int  enetLine = -1;
static int  enetProff,enetSblock,enetPage,enetCmxfcrMask;
static int  enetCmxfcrValue,enetCpxfcrRamType,enetPsmr;

extern	int 	mii_init_phy ( int phyNr );

/*******************************************************************************
*
* fccSelect - select fcc ethernet interface
*
* NOTES: N/A
*
* ARGUMENTS:  N?A
*
* LIMITATIONS: N/A 
*
* RETURNS: N/A
*
* SEE ALSO:
*/
int	fccSelect ( int fccNr )
{
if ( enetLine==fccNr ) return(0);

switch (fccNr)
	{
#ifdef CFG_ETH_FCC1
	case 1 :
		enetProff = PROFF_FCC1;
		enetSblock = CPM_CR_FCC1_SBLOCK;
		enetPage = CPM_CR_FCC1_PAGE;
		enetCmxfcrMask = CFG_ETH1_CMXFCR_MASK;
		enetCmxfcrValue = CFG_ETH1_CMXFCR_VALUE;
		enetCpxfcrRamType = CFG_ETH1_CPMFCR_RAMTYPE;
		enetPsmr = CFG_ETH1_FCC_PSMR;
		break;
#endif
#ifdef CFG_ETH_FCC2
	case 2 :
		enetProff = PROFF_FCC2;
		enetSblock = CPM_CR_FCC2_SBLOCK;
		enetPage = CPM_CR_FCC2_PAGE;
		enetCmxfcrMask = CFG_ETH2_CMXFCR_MASK;
		enetCmxfcrValue = CFG_ETH2_CMXFCR_VALUE;
		enetCpxfcrRamType = CFG_ETH2_CPMFCR_RAMTYPE;
		enetPsmr = CFG_ETH2_FCC_PSMR;
		break;
#endif
#ifdef CFG_ETH_FCC3
	case 3 :
		enetProff = PROFF_FCC3;
		enetSblock = CPM_CR_FCC3_SBLOCK;
		enetPage = CPM_CR_FCC3_PAGE;
		enetCmxfcrMask = CFG_ETH3_CMXFCR_MASK;
		enetCmxfcrValue = CFG_ETH3_CMXFCR_VALUE;
		enetCpxfcrRamType = CFG_ETH3_CPMFCR_RAMTYPE;
		enetPsmr = CFG_ETH3_FCC_PSMR;
		break;
#endif
	default  :
		printf ("ERROR - FCC eth interface unknown (%d)\n",fccNr);
		return (-1);
	}
enetLine=fccNr;
#ifdef DEBUG
printf("select fcc %d\n",fccNr);
#endif
return (0);
}		

/*******************************************************************************
*
* fcc_send - send fcc ethernet interface
*
* NOTES: N/A
*
* ARGUMENTS:  N?A
*
* LIMITATIONS: N/A 
*
* RETURNS: N/A
*
* SEE ALSO:
*/
static int fcc_send(struct eth_device* dev, volatile void *packet, int length)
{
int 	i;
int 	result = 0;

if ( fccSelect(DEV_FCC_NR)!=0 ) goto out;

if (length <= 0) 
	{
	printf("fcc %d : bad packet size: %d\n",DEV_FCC_NR, length);
	goto out;
    	}
for (i=0; rtx.txbd[txIdx[DEV_FCC_NR-1]].cbd_sc & BD_ENET_TX_READY; i++) 
	{
	if (i >= TOUT_LOOP) 
		{
	    	printf("fcc %d : tx buffer not ready\n",DEV_FCC_NR);
	    	goto out;
		}
    	}
rtx.txbd[txIdx[DEV_FCC_NR-1]].cbd_bufaddr = (uint)packet;
rtx.txbd[txIdx[DEV_FCC_NR-1]].cbd_datlen = length;
rtx.txbd[txIdx[DEV_FCC_NR-1]].cbd_sc |= (BD_ENET_TX_READY | BD_ENET_TX_LAST | BD_ENET_TX_WRAP);

for(i=0; rtx.txbd[txIdx[DEV_FCC_NR-1]].cbd_sc & BD_ENET_TX_READY; i++) 
	{
	if (i >= TOUT_LOOP) 
		{
	    	printf("fcc %d : tx error\n",DEV_FCC_NR);
	    	goto out;
		}
    	}

/* return only status bits */
result = rtx.txbd[txIdx[DEV_FCC_NR-1]].cbd_sc & BD_ENET_TX_STATS;

#ifdef DEBUG
printf("fcc %d - send frame - txbd status: %04x - result: %04x\n",
	DEV_FCC_NR, rtx.txbd[txIdx[DEV_FCC_NR-1]].cbd_sc, result);
#endif

out:
return (result);
}

/*******************************************************************************
*
* fcc_recv - receive fcc ethernet interface
*
* NOTES: N/A
*
* ARGUMENTS:  N?A
*
* LIMITATIONS: N/A 
*
* RETURNS: N/A
*
* SEE ALSO:
*/
int	fcc_recv (struct eth_device* dev)
{
int 	length;

if ( fccSelect(DEV_FCC_NR)!=0 ) return(0);

for (;;)
    	{
	if (rtx.rxbd[rxIdx[DEV_FCC_NR-1]].cbd_sc & BD_ENET_RX_EMPTY) 
		{
	    	length = -1;
	    	break;     /* nothing received - leave for() loop */
		}
	length = rtx.rxbd[rxIdx[DEV_FCC_NR-1]].cbd_datlen;

	if (rtx.rxbd[rxIdx[DEV_FCC_NR-1]].cbd_sc & 0x003f) 
		{
	    	printf("fcc %d : rx error %04x\n",DEV_FCC_NR, rtx.rxbd[rxIdx[DEV_FCC_NR-1]].cbd_sc);
		}
	else 	{
#ifdef DEBUG
	    	printf("fcc %d : rx frame - len=%d\n",DEV_FCC_NR, length);
#endif
	    	/* Pass the packet up to the protocol layers. */
	    	NetReceive(NetRxPackets[rxIdx[DEV_FCC_NR-1]], length - 4);
		}
	/* Give the buffer back to the FCC. */
	rtx.rxbd[rxIdx[DEV_FCC_NR-1]].cbd_datlen = 0;
	/* wrap around buffer index when necessary */
	if ((rxIdx[DEV_FCC_NR-1] + 1) >= PKTBUFSRX) 
		{
	    	rtx.rxbd[PKTBUFSRX - 1].cbd_sc = (BD_ENET_RX_WRAP | BD_ENET_RX_EMPTY);
	    	rxIdx[DEV_FCC_NR-1] = 0;
		}
	else 	{
	    	rtx.rxbd[rxIdx[DEV_FCC_NR-1]].cbd_sc = BD_ENET_RX_EMPTY;
	    	rxIdx[DEV_FCC_NR-1]++;
		}
    	}
return length;
}

/*******************************************************************************
*
* fcc_init - init fcc ethernet interface
*
* NOTES: N/A
*
* ARGUMENTS:  N?A
*
* LIMITATIONS: N/A 
*
* RETURNS: N/A
*
* SEE ALSO:
*/
int fcc_init(struct eth_device* dev, bd_t *bis)
{
int i;
volatile immap_t *immr = (immap_t *)CFG_IMMR;
volatile cpm8260_t *cp = &(immr->im_cpm);
fcc_enet_t *pram_ptr;
unsigned long mem_addr;

if ( fccSelect(DEV_FCC_NR)!=0 ) return(0);

#ifdef DEBUG
 printf("fcc %d : init\n",DEV_FCC_NR);
#endif

/* 28.9 - (1-2): ioports have been set up already */

/* 28.9 - (3): connect FCC's tx and rx clocks */
immr->im_cpmux.cmx_uar = 0;
immr->im_cpmux.cmx_fcr = (immr->im_cpmux.cmx_fcr & ~enetCmxfcrMask) | enetCmxfcrValue;
/* 28.9 - (4): GFMR: disable tx/rx, CCITT CRC, Mode Ethernet */
immr->im_fcc[DEV_FCC_NR-1].fcc_gfmr = FCC_GFMR_MODE_ENET | FCC_GFMR_TCRC_32;
/* 28.9 - (5): FPSMR: enable full duplex, select CCITT CRC for Ethernet */
immr->im_fcc[DEV_FCC_NR-1].fcc_fpsmr = enetPsmr | FCC_PSMR_ENCRC;
/* 28.9 - (6): FDSR: Ethernet Syn */
immr->im_fcc[DEV_FCC_NR-1].fcc_fdsr = 0xD555;
/* reset indeces to current rx/tx bd (see fcc_send()/fcc_rx()) */
rxIdx[DEV_FCC_NR-1] = 0;
txIdx[DEV_FCC_NR-1] = 0;
/* Setup Receiver Buffer Descriptors */
for (i = 0; i < PKTBUFSRX; i++)
	{
	rtx.rxbd[i].cbd_sc = BD_ENET_RX_EMPTY;
	rtx.rxbd[i].cbd_datlen = 0;
	rtx.rxbd[i].cbd_bufaddr = (uint)NetRxPackets[i];
	}
rtx.rxbd[PKTBUFSRX - 1].cbd_sc |= BD_ENET_RX_WRAP;
/* Setup Ethernet Transmitter Buffer Descriptors */
for (i = 0; i < TX_BUF_CNT; i++)
	{
	rtx.txbd[i].cbd_sc = (BD_ENET_TX_PAD | BD_ENET_TX_LAST | BD_ENET_TX_TC);
	rtx.txbd[i].cbd_datlen = 0;
	rtx.txbd[i].cbd_bufaddr = (uint)&txbuf[i][0];
	}
rtx.txbd[TX_BUF_CNT - 1].cbd_sc |= BD_ENET_TX_WRAP;
/* 28.9 - (7): initialise parameter ram */
pram_ptr = (fcc_enet_t *)&(immr->im_dprambase[enetProff]);
/* clear whole structure to make sure all reserved fields are zero */
memset((void*)pram_ptr, 0, sizeof(fcc_enet_t));
/*
 * common Parameter RAM area
 *
 * Allocate space in the reserved FCC area of DPRAM for the
 * internal buffers.  No one uses this space (yet), so we
 * can do this.  Later, we will add resource management for
 * this area.
 */
mem_addr = CPM_FCC_SPECIAL_BASE + ((DEV_FCC_NR-1) * 64);
pram_ptr->fen_genfcc.fcc_riptr = mem_addr;
pram_ptr->fen_genfcc.fcc_tiptr = mem_addr+32;
/* Set maximum bytes per receive buffer. It must be a multiple of 32. */
pram_ptr->fen_genfcc.fcc_mrblr = PKT_MAXBLR_SIZE;
pram_ptr->fen_genfcc.fcc_rstate = (CPMFCR_GBL | CPMFCR_EB | enetCpxfcrRamType) << 24;
pram_ptr->fen_genfcc.fcc_rbase = (unsigned int)(&rtx.rxbd[rxIdx[DEV_FCC_NR-1]]);
pram_ptr->fen_genfcc.fcc_tstate = (CPMFCR_GBL | CPMFCR_EB | enetCpxfcrRamType) << 24;
pram_ptr->fen_genfcc.fcc_tbase = (unsigned int)(&rtx.txbd[txIdx[DEV_FCC_NR-1]]);
/* protocol-specific area */
pram_ptr->fen_cmask = 0xdebb20e3;	/* CRC mask */
pram_ptr->fen_cpres = 0xffffffff;	/* CRC preset */
pram_ptr->fen_retlim = 15;		/* Retry limit threshold */
pram_ptr->fen_mflr = PKT_MAXBUF_SIZE;   /* maximum frame length register */
/* Set Ethernet station address. */
#define ea dev->enetaddr 
pram_ptr->fen_paddrh = (ea[5] << 8) + ea[4];
pram_ptr->fen_paddrm = (ea[3] << 8) + ea[2];
pram_ptr->fen_paddrl = (ea[1] << 8) + ea[0];
#undef ea
pram_ptr->fen_minflr = PKT_MINBUF_SIZE; /* minimum frame length register */
/* pad pointer. use tiptr since we don't need a specific padding char */
pram_ptr->fen_padptr = pram_ptr->fen_genfcc.fcc_tiptr;
pram_ptr->fen_maxd1 = PKT_MAXDMA_SIZE;	/* maximum DMA1 length */
pram_ptr->fen_maxd2 = PKT_MAXDMA_SIZE;	/* maximum DMA2 length */
pram_ptr->fen_rfthr = 1;
pram_ptr->fen_rfcnt = 1;
/*
#ifdef DEBUG
printf("pram_ptr->fen_genfcc.fcc_rbase %08x\n", pram_ptr->fen_genfcc.fcc_rbase);
printf("pram_ptr->fen_genfcc.fcc_tbase %08x\n", pram_ptr->fen_genfcc.fcc_tbase);
#endif
*/
/* 28.9 - (8): clear out events in FCCE */
immr->im_fcc[DEV_FCC_NR-1].fcc_fcce = ~0x0;
/* 28.9 - (9): FCCM: mask all events */
immr->im_fcc[DEV_FCC_NR-1].fcc_fccm = 0;
/* 28.9 - (10-12): we don't use ethernet interrupts */
/* 28.9 - (13)
*
* Let's re-initialize the channel now.  We have to do it later
* than the manual describes because we have just now finished
* the BD initialization.
*/
cp->cp_cpcr = mk_cr_cmd(enetPage, enetSblock, 0x0c, CPM_CR_INIT_TRX) | CPM_CR_FLG;
do 	{
	__asm__ __volatile__ ("eieio");
    	} while (cp->cp_cpcr & CPM_CR_FLG);
/* 28.9 - (14): enable tx/rx in gfmr */
immr->im_fcc[DEV_FCC_NR-1].fcc_gfmr |= FCC_GFMR_ENT | FCC_GFMR_ENR;

return (1);
}

/*******************************************************************************
*
* fcc_halt - halt fcc ethernet interface
*
* NOTES: N/A
*
* ARGUMENTS:  N?A
*
* LIMITATIONS: N/A 
*
* RETURNS: N/A
*
* SEE ALSO:
*/
static void fcc_halt(struct eth_device* dev)
{
volatile immap_t *immr = (immap_t *)CFG_IMMR;

if ( fccSelect(DEV_FCC_NR)!=0 ) return;

#ifdef DEBUG
 printf("fcc %d : halt\n",DEV_FCC_NR);
#endif

/* write GFMR: disable tx/rx */
immr->im_fcc[DEV_FCC_NR-1].fcc_gfmr &= ~(FCC_GFMR_ENT | FCC_GFMR_ENR);
}

/*******************************************************************************
*
* getMACAddr - get the eth???addr with value from RI
*
* NOTES: 	if RI is not available use the dafault value
*
* ARGUMENTS:  ethernet number, pointer to macAddr (6 bytes)
*
* LIMITATIONS: ethnR = 0:2
*
* RETURNS: N/A
*
* SEE ALSO:
*/
void 	getMACAddr (int ethNr, unsigned char *macAddr )
{
uchar	RIethAddr[6];
char	ethA[20];
char	*ethAddr = ethA;
char	*tmp, *end;
int	i;
int	result = -1;

#ifdef CONFIG_RINV
 /* get MAC from RI */
 if ( rinvCheck(CFG_I2C_SLAVE_MB,0)==0 ) 
 	 {
 	 if ( rinvGetMACAddress ( CFG_I2C_SLAVE_MB, (uchar *) 0, ethNr, RIethAddr ) == 0 )	     
 		 {
 		 /* controllo se MAC Italtel */
 		 if ( RIethAddr[0]==0x0 && RIethAddr[1]==0x1 && RIethAddr[2]==0x5b )
 			 {
 			 sprintf ( ethAddr, "%02x:%02x:%02x:%02x:%02x:%02x", 
 				 RIethAddr[0], RIethAddr[1], RIethAddr[2], 
 				 RIethAddr[3], RIethAddr[4], RIethAddr[5] );
 			 result = 0;
 			 }
 		 else	 printf ("\n*** Warning - bad eth %d address from RI",ethNr);
 		 }
 	 else	 printf ("\n*** Warning - error reading eth %d address from RI",ethNr);
 	 }
#endif

if ( result!=0 ) 
	{
	/* set MAC default value */
	switch (ethNr)
		{
		#ifdef CONFIG_ETHADDR
		case 0: if ( (ethAddr=getenv("ethaddr"))==NULL ) ethAddr = MK_STR(CONFIG_ETHADDR);
			break;
		#endif
		#ifdef CONFIG_ETHADDR1
		case 1: if ( (ethAddr=getenv("eth1addr"))==NULL ) ethAddr = MK_STR(CONFIG_ETHADDR1);
			break;
		#endif
		#ifdef CONFIG_ETHADDR2
		case 2: if ( (ethAddr=getenv("eth2addr"))==NULL ) ethAddr = MK_STR(CONFIG_ETHADDR2);
			break;
		#endif
		default:printf ("\n    ERROR - default eth %d address not defined\n", ethNr);  
			return;
		}
	printf (" - using default ( ");
	tmp = ethAddr;
	for (i=0; i<6; i++) 
		{
		RIethAddr[i] = tmp ? simple_strtoul(tmp, &end, 16) : 0;
		if (tmp) tmp = (*end) ? end+1 : end;
		}
	for (i=0; i<5; i++) printf ("%x:",RIethAddr[i]);
	printf ("%x )\n",RIethAddr[i]);
	}

#ifdef DEBUG
printf (">> eth%d - %s <<\n",ethNr, ethAddr );
#endif

if ( macAddr!=NULL ) memcpy(macAddr, RIethAddr, 6);
}

/*******************************************************************************
*
* setMACAddr - set the enviroment variable eth???addr 
*
* NOTES: N/A	
*
* ARGUMENTS:  ethernet number,pointer to macAddr (6 bytes)
*
* LIMITATIONS: ethnR = 0:2
*
* RETURNS: N/A
*
* SEE ALSO:
*/
void 	setMACAddr (int ethNr, unsigned char *macAddr )
{
char	ethAddr[20];

sprintf ( ethAddr, "%02x:%02x:%02x:%02x:%02x:%02x", 
	macAddr[0], macAddr[1], macAddr[2], 
	macAddr[3], macAddr[4], macAddr[5] );

/* set MAC value */
switch (ethNr)
	{
	case 0: setenv("ethaddr",ethAddr);
		break;
	case 1: setenv("eth1addr",ethAddr);
		break;
	case 2: setenv("eth2addr",ethAddr);
		break;
	default: break;
	}
}

/*******************************************************************************
*
* fec_init - init fcc ethernet interface
*
* NOTES: N/A
*
* ARGUMENTS:  N?A
*
* LIMITATIONS: N/A 
*
* RETURNS: N/A
*
* SEE ALSO:
*/
void 	fec_init(int ethNr, int fccNr, bd_t *bis)
{
struct 	eth_device *dev;

dev = (struct eth_device*) malloc(sizeof *dev);
sprintf(dev->name, "FCC %d", fccNr);
dev->iobase = 0;
dev->init   = fcc_init;
dev->halt   = fcc_halt;
dev->send   = fcc_send;
dev->recv   = fcc_recv;
/* get eth address (MAC) */
getMACAddr(ethNr,dev->enetaddr);
DEV_FCC_NR = fccNr;
eth_register(dev);
#ifdef DEBUG
 printf ("register eth %d on FCC %d\n",ethNr,fccNr);
 printf("\tMAC = %02X:%02X:%02X:%02X:%02X:%02X\n",
       dev->enetaddr[0], dev->enetaddr[1],dev->enetaddr[2], dev->enetaddr[3],
       dev->enetaddr[4], dev->enetaddr[5]);
#endif
setMACAddr(ethNr,dev->enetaddr); 
}

/*******************************************************************************
*
* fec_initialize - init all fcc ethernet interface
*
* NOTES: N/A
*
* ARGUMENTS:  N?A
*
* LIMITATIONS: N/A 
*
* RETURNS: N/A
*
* SEE ALSO:
*/
int 	fec_initialize(bd_t *bis)
{
int	ethNr = 0;

#ifdef CFG_ETH_FCC1
 fec_init(ethNr,1,bis); 
 #ifdef CONFIG_MII
  mii_init_phy(ethNr);
 #endif
 ethNr++;
#endif
#ifdef CFG_ETH_FCC2
 fec_init(ethNr,2,bis);
 #ifdef CONFIG_MII
  mii_init_phy(ethNr);
 #endif
 ethNr++;
#endif
#ifdef CFG_ETH_FCC3
 fec_init(ethNr,3,bis);
 #ifdef CONFIG_MII
  mii_init_phy(ethNr);
 #endif
#endif

return 1;
}

/******************************************************************************/
#endif	/* FEC_ENET && CFG_CMD_NET */
